import wx
from wx.grid import (
    GridCellAttr,
    GridTableMessage,
    GRIDTABLE_NOTIFY_ROWS_DELETED,
    GRIDTABLE_NOTIFY_ROWS_APPENDED,
)


class SchGridTableBase(wx.grid.GridTableBase):
    def __init__(self):
        wx.grid.GridTableBase.__init__(self)
        self.can_append = 0
        self.append_count = 0
        self.read_only = True
        self.rec_to_update = dict()
        self.rec_to_instert = dict()
        self.rec_to_delete = []
        self.rec_selected = []
        self.tabsort = []
        self.filter_id = None
        self.key = None
        self.sel_mask = []

        self.sel_colour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_INFOBK)
        self.attr_insert = GridCellAttr()
        self.attr_insert.SetBackgroundColour("PALE GREEN")
        self.attr_update = GridCellAttr()
        self.attr_update.SetBackgroundColour("WHEAT")
        self.attr_del = GridCellAttr()
        self.attr_del.SetBackgroundColour("RED")
        self.attr_del.SetFont(wx.Font(8, wx.ROMAN, wx.ITALIC, wx.NORMAL, True))
        self.attr_sel = GridCellAttr()
        self.attr_sel.SetBackgroundColour(self.sel_colour)
        self.attr_sel.SetFont(wx.Font(8, wx.ROMAN, wx.ITALIC, wx.NORMAL, True))
        self.attr_normal = GridCellAttr()

        self.data = []
        self.data_org = None
        self.last_row_count = 0
        self.auto_size = "simple"
        self.no_actions = False

        self.init_data_base()

    def init_data_base(self):
        self.append_count = 0
        self.read_only = True
        self.rec_to_update = dict()
        self.rec_to_instert = dict()
        self.rec_to_delete = []
        self.rec_selected = []
        self.data = []
        self.data_org = None
        self.last_row_count = 0

    def replace_tab(self, new_tab):
        c = self.last_row_count
        # self.refr_count(0)
        self.init_data_base()
        self.data = new_tab
        self.last_row_count = c

    def refr_count(self, count, store_pos=0):
        if count == self.last_row_count:
            return
        old_count = self.last_row_count
        self.last_row_count = count
        self.GetView().refr_count(old_count, count, store_pos)

    def _get_number_rows(self):
        return len(self.data) + self.can_append + self.append_count

    def get_rec(self, row):
        try:
            return self.data[row]
        except IndexError:
            return []

    def get_sort_nr(self, col):
        c = col + 1
        if c in self.tabsort:
            return self.tabsort.index(c) + 1
        else:
            if -1 * c in self.tabsort:
                return -1 * self.tabsort.index(-1 * c) - 1
            else:
                return 0

    def sort(self, column, append):
        col = column + 1
        if append:
            if col in self.tabsort:
                self.tabsort.remove(col)
                self.tabsort.append(-1 * col)
            else:
                if -1 * col in self.tabsort:
                    self.tabsort.remove(-1 * col)
                else:
                    self.tabsort.append(col)
        else:
            if col in self.tabsort:
                self.tabsort = [-1 * col]
            else:
                if -1 * col in self.tabsort:
                    self.tabsort = []
                else:
                    self.tabsort = [col]

    def _is_sel(self, row):
        if row in self.rec_selected:
            return True
        return False

    def sel_row(self, row):
        if row < self.GetNumberRows():
            if row in self.rec_selected:
                self.rec_selected.remove(row)
            else:
                self.rec_selected.append(row)
            self.GetView().ForceRefresh()
            return True
        else:
            return False

    def make_exec_cmd(self, cmd, option=0):
        if option == 0:
            data = [cmd]
        elif option == 1:
            selrows = self.get_sel_rows()
            key = self.get_sel_key()
            data = [cmd, (key,) + selrows]
        else:
            data = [cmd]
        return data

    def run_cmd(self, cmd, options):
        pass

    def get_sel_key(self):
        row = self.GetView().GetGridCursorRow()
        rec = self.get_rec(row)
        if rec:
            return rec[0]
        else:
            return None

    def get_sel_rows(self):
        ret = []
        for pos in self.rec_selected:
            ret.append(self.get_rec(pos)[0])
        return (ret, self.sel_mask, self.key)

    def set_read_only(self, read_only):
        self.read_only = read_only
        if self.read_only:
            if self.can_append == 1:
                self.last_row_count -= 1
            self.can_append = 0
        else:
            if self.can_append == 0:
                self.last_row_count += 1
            self.can_append = 1

    def set_no_actions(self, val):
        self.no_actions = val

    def is_read_only(self, row, col):
        return self.read_only

    def filter_cmp(self, pos, key):
        if self.filter_id is not None:
            if str(pos[self.filter_id]).startswith(key):
                return True
            else:
                return False
        else:
            return False

    def filter_by(self, col):
        self.filter_id = col

    def filter(self, key):
        if self.key and self.key == key:
            return
        old_count = self.GetNumberRows()
        if key and len(key) > 0:
            if not self.data_org:
                self.data_org = self.data
            self.data = []
            for pos in self.data_org:
                if self.filter_cmp(pos, key):
                    self.data.append(pos)
        else:
            if self.data_org:
                self.data = self.data_org
                self.data_org = None
        self.key = key
        self.refr_count(old_count, self.GetNumberRows(), 0)
        self.GetView().ForceRefresh()

    def append_sel_mask(self, mask):
        self.sel_mask.append(mask)
        self.GetView().ForceRefresh()

    def clear_sel_masks(self):
        self.sel_mask = []

    def clear_state(self):
        self.append_count = 0
        self.rec_to_update = dict()
        self.rec_to_instert = dict()
        self.rec_to_delete = []
        self.rec_selected = []
        self.tabsort = []
        self.filter_id = None
        self.key = None
        self.sel_mask = []

    def get_table_and_state(self):
        tableandstate = []
        tableandstate.append(
            (
                self.append_count,
                self.rec_to_update,
                self.rec_to_instert,
                self.rec_to_delete,
                self.rec_selected,
                self.tabsort,
                self.filter_id,
                self.key,
                self.sel_mask,
                self.last_row_count,
            )
        )
        return tableandstate

    def duplicate_table_and_state(self):
        tableandstate = self.get_table_and_state()
        self.append_count = 0
        self.rec_to_update = dict()
        self.rec_to_instert = dict()
        self.rec_to_delete = []
        self.rec_selected = []
        self.tabsort = []
        self.filter_id = None
        self.key = None
        self.sel_mask = []
        return tableandstate

    def set_table_and_state(self, tableandstate):
        oldtableandstate = self.get_table_and_state()
        rec = tableandstate[0]
        self.append_count = rec[0]
        self.rec_to_update = rec[1]
        self.rec_to_instert = rec[2]
        self.rec_to_delete = rec[3]
        self.rec_selected = rec[4]
        self.tabsort = rec[5]
        self.filter_id = rec[6]
        self.key = rec[7]
        self.sel_mask = rec[8]
        self.last_row_count = rec[9]
        return oldtableandstate

    def get_actions(self, row, col=None):
        ret = {}
        attrs = self.get_action_list(row, col)
        for attr in attrs:
            if "name" in attr:
                name = attr["name"]
                if "href" in attr:
                    href = attr["href"]
                    if "title" in attr:
                        title = attr["title"]
                    else:
                        title = ""
                    ret[name] = ("openurl", href, title, attr)
        if len(ret) > 0:
            return ret
        return None

    def enable(self, en):
        pass

    def GetNumberRows(self):
        count = self._get_number_rows()
        return count

    def GetNumberCols(self):
        if len(self.data) > 0:
            return len(self.data[0]) - 1
        else:
            return 0

    def IsEmptyCell(self, row, col):
        try:
            return not self.data[row][col]
        except IndexError:
            return True

    def GetValue(self, row, col):
        try:
            return self.get_rec(row)[col]
        except IndexError:
            return ""

    def GetColLabelValue(self, col):
        return "COL_" + str(col)

    def GetTypeName(self, row, col):
        return self.GetValue(row, col).__class__.__name__

    def CanGetValueAs(self, row, col, type_name):
        if self.GetTypeName(row, col) == type_name:
            return True
        else:
            return False

    def CanSetValueAs(self, row, col, type_name):
        return self.CanGetValueAs(row, col, type_name)

    def GetAttr(self, row, col, kind):
        if row < (self.GetNumberRows() - self.can_append) - self.append_count:
            if row in self.rec_to_update:
                attr = self.attr_update
                attr.IncRef()
                return attr
            elif row in self.rec_to_delete:
                attr = self.attr_del
                attr.IncRef()
                return attr
            elif self._is_sel(row):
                attr = self.attr_sel
                attr.IncRef()
            else:
                attr = self.attr_normal
                attr.IncRef()
            return attr
        else:
            attr = self.attr_insert
            attr.IncRef()
            return attr

    def refresh(self, storePos):
        pass

    def copy(self):
        return None

    def paste(self, data):
        return None



class Signal:
    """Helper class for sending signals between windows"""

    def __init__(self):
        self._signals = {}

    def register_signal(self, obj, signal_name):
        """Register obj to receive signal.
        Object must have a function with name: signal_name
        Function must accept arguments passed to method: signal

        Args:
            obj - object to receive signals
            signal_name - name of signal
        """
        if signal_name not in self._signals:
            self._signals[signal_name] = []
        if not obj in self._signals[signal_name]:
            self._signals[signal_name].append(obj)

    def unregister_signal(self, obj, signal_name):
        """Unregister obj to receive signal

        Args:
            obj - object to receive signals
            signal_name - name of signal
        """
        if obj in self._signals[signal_name]:
            id = self._signals[signal_name].index(obj)
            del self._signals[signal_name][id]

    def signal(self, signal_name, *argi, **argv):
        """Send signal

        Args:
            signal_name - name of signal
            *argi, **argv - parameters of signal
        """
        ret = []
        if signal_name in self._signals:
            for obj in self._signals[signal_name]:
                x = getattr(obj, signal_name)(*argi, **argv)
                if x is not None:
                    ret.append(x)
        return ret



"""Module contains all widget class which can be used on SchForm"""

from base64 import b64encode
import os
import platform
import datetime

import wx

from wx import ComboCtrl
from wx.adv import BitmapComboBox, CalendarCtrl, EditableListBox, DatePickerCtrl
from wx.lib.agw.hypertreelist import HyperTreeList as TreeListCtrl
from wx.lib import ticker, masked, colourselect, filebrowsebutton

import wx.lib.imagebrowser
import wx.lib.platebtn as platebtn
import wx.lib.buttons as buttons

from pytigon_lib.schtools import createparm
from pytigon_lib.schparser.html_parsers import ShtmlParser

from pytigon_lib.schhtml.wxdc import DcDc
from pytigon_lib.schhtml.htmlviewer import HtmlViewerParser, tdata_from_html
from pytigon_lib.schparser.html_parsers import Td
from pytigon_lib.schtools.tools import is_null

from pytigon_gui.guictrl.grid import grid, gridtable_from_proxy, tabproxy
from pytigon_gui.guiframe import page
from pytigon_gui.guilib.image import bitmap_from_href

from pytigon_gui.guictrl.grid.gridtable_from_html_table import SimpleDataTable
from pytigon_gui.guictrl.grid.gridpanel import SchGridPanel
from pytigon_gui.guictrl.popup.popuphtml import DataPopupControl
from pytigon_gui.guictrl.popup.select2 import Select2Base
from pytigon_gui.guictrl.basectrl import SchBaseCtrl, handle_best_size
from pytigon_gui.guictrl.button.toolbarbutton import BitmapTextButton

from django.utils.translation import gettext_lazy as _


def SELECT(parent, **kwds):
    """SELECT handle ctrlselect tag

    Tag arguments:
        multiple -
    """
    size = 1
    multiple = False

    if "param" in kwds:
        param = kwds["param"]
        if "multiple" in param:
            multiple = True
        if "size" in param:
            size = int(param["size"])
        if multiple and size == 1:
            size = 10

        if "schtype" in param and param["schtype"] in (
            "HeavySelect2Field",
            "HeavySelect2MultipleField",
            "ModelSelect2Field",
            "ModelSelect2MultipleField",
        ):
            kwds["param"]["data"] = [
                kwds["param"],
            ]
            kwds["param"]["data"][0]["attrs"] = kwds["param"]

            return SELECT2(parent, **kwds)
    else:
        size = 1

    if size == 1:
        kwds["readonly"] = True
        return DBCHOICE(parent, **kwds)
    else:
        kwds["length"] = size
        #        return HTMLLISTBOX(parent, **kwds)
        # return CHECKLIST(parent, **kwds)
        return CHECKLISTBOX(parent, **kwds)


def BUTTON(parent, **kwds):
    """BUTTON handle ctrlbutton tag

    Tag arguments:
        src - link to button bitmap

    Subclass:
        SIMPLE_BUTTON
        BITMAPBUTTON
        PLATEBUTTON
        GENBITMAPBUTTON
        GENBITMAPTEXTBUTTON
        GENBITMAPBUTTONTXT
        GENBITMAPBUTTONTXT_SMALL
        NOBG_BUTTON
        NOBG_BUTTON_TXT
        CLOSEBUTTON
        MENUBUTTON
        MENUTOOLBARBUTTON
    """
    if "param" in kwds and "btn-class" in kwds["param"]:
        c = kwds["param"]["btn-class"]
        if c == "SIMPLE_BUTTON":
            return SIMPLE_BUTTON(parent, **kwds)
        elif c == "BITMAPBUTTON":
            return BITMAPBUTTON(parent, **kwds)
        elif c == "PLATEBUTTON":
            return PLATEBUTTON(parent, **kwds)
        elif c == "GENBITMAPBUTTON":
            return GENBITMAPBUTTON(parent, **kwds)
        elif c == "GENBITMAPBUTTONTXT":
            return GENBITMAPBUTTONTXT(parent, **kwds)
        elif c == "GENBITMAPBUTTONTXT_SMALL":
            return GENBITMAPBUTTONTXT_SMALL(parent, **kwds)
        elif c == "NOBG_BUTTON":
            return NOBG_BUTTON(parent, **kwds)
        elif c == "NOBG_BUTTON_TXT":
            return NOBG_BUTTON_TXT(parent, **kwds)
        elif c == "CLOSEBUTTON":
            return CLOSEBUTTON(parent, **kwds)
        elif c == "MENUBUTTON":
            return MENUBUTTON(parent, **kwds)
        elif c == "MENUTOOLBARBUTTON":
            return MENUTOOLBARBUTTON(parent, **kwds)
    if "src" in kwds:
        ctrl = BITMAPBUTTON(parent, **kwds)
    else:
        ctrl = SIMPLE_BUTTON(parent, **kwds)
    return ctrl


def _make_button_class(
    base_class, is_bitmap_button=False, is_close_button=False, icon_size=2
):
    class BUTTONCLASS(base_class, SchBaseCtrl):
        def __init__(self, parent, **kwds):
            SchBaseCtrl.__init__(self, parent, kwds)
            if is_bitmap_button:
                self._set_bitmap()
                if (
                    base_class == buttons.GenBitmapButton
                    or base_class == BitmapTextButton
                ):
                    if "style" in kwds:
                        style = kwds["style"] | wx.NO_BORDER
                        kwds["style"] = style
                    else:
                        style = wx.NO_BORDER
                        kwds["style"] = wx.NO_BORDER
                    kwds["bitmap"] = self.bmp
                elif base_class == buttons.GenBitmapTextButton:
                    kwds["bitmap"] = self.bmp
                elif base_class == platebtn.PlateButton:
                    kwds["style"] = (
                        platebtn.PB_STYLE_SQUARE | platebtn.PB_STYLE_GRADIENT
                    )
                    kwds["bmp"] = self.bmp
                else:
                    if "style" in kwds:
                        style = kwds["style"]
                    else:
                        kwds["style"] = 0
                        style = 0
                    kwds["bitmap"] = self.bmp

                if base_class != wx.BitmapButton:
                    if self.label:
                        kwds["label"] = self.label.replace("&", "")
            else:
                if "style" not in kwds:
                    kwds["style"] = 0
                    style = 0
                if self.label:
                    kwds["label"] = self.label.replace("&", "")
                if self.nr_id:
                    try:
                        nr_id = int(self.nr_id)
                    except:
                        if "wx." in self.nr_id:
                            nr_id = eval(self.nr_id)
                        else:
                            nr_id = None
                    if nr_id != None:
                        kwds["id"] = nr_id

            if "fields" in kwds:
                self.fields = kwds["fields"]
                del kwds["fields"]
            else:
                self.fields = None

            base_class.__init__(self, parent, **kwds)

            if self.defaultvalue:
                self.SetDefault()
                self.parent.any_parent_command("set_default_button", self)

            if is_bitmap_button and self.label:
                self.SetToolTip(self.label.replace("&", ""))

            if is_close_button:
                self.Bind(wx.EVT_BUTTON, self.on_exit)
            else:
                self.Bind(wx.EVT_BUTTON, self.on_click, self)

            if self.label and base_class == BitmapTextButton:
                self.SetToolTip(wx.ToolTip(self.label.replace("&", "")))

            if self.label and "&" in self.label:
                id = self.label.find("&")
                l = self.label[id + 1 : id + 2]
                if l:
                    aTable = [
                        (wx.ACCEL_ALT, ord("N"), self.on_acc_click),
                    ]
                    self.set_acc_key_tab(aTable)

            self.init2()

        def on_acc_click(self, event):
            evt = wx.PyCommandEvent(wx.EVT_BUTTON.typeId, self.GetId())
            wx.PostEvent(self, evt)

        def _set_bitmap(self):
            self.bmp = None
            if is_bitmap_button:
                if is_close_button:
                    self.src = "png://emblems/emblem-unreadable.png"
                if self.src:
                    self.bmp = bitmap_from_href(self.src, icon_size)

        def process_refr_data(self, **kwds):
            self.init_base(kwds)

            if is_bitmap_button:
                self._set_bitmap()

                if base_class == platebtn.PlateButton:
                    kwds["bmp"] = self.bmp
                else:
                    kwds["bitmap"] = self.bmp

            if "fields" in kwds:
                self.fields = kwds["fields"]
                del kwds["fields"]
            else:
                self.fields = None

            if "href" in kwds:
                self.href = kwds["href"]
                del kwds["href"]

            if self.defaultvalue:
                self.SetDefault()

            self.init2()

        def init2(self):
            if self.href == "_hide":
                self.Hide()
            else:
                self.Show()
            if self.href == "_disable":
                self.Enable(False)
            else:
                self.Enable(True)
            if not self.style:
                self.style = 0

        def CanAcceptFocus(self):
            return False

        if is_close_button:

            def on_exit(self, event):
                def fun():
                    self.GetParent().any_parent_command("on_child_form_cancel")

                wx.CallAfter(fun)

        else:

            def on_click(self, event):
                upload = False
                if self.valuetype == "upload":
                    upload = True
                if hasattr(self, "href"):
                    href = self.href
                else:
                    href = ""
                self.get_parent_form().href_clicked(
                    self, {"href": href, "target": self.target}, upload, self.fields
                )

    return BUTTONCLASS


SIMPLE_BUTTON = _make_button_class(wx.Button)
BITMAPBUTTON = _make_button_class(wx.BitmapButton, is_bitmap_button=True)
PLATEBUTTON = _make_button_class(platebtn.PlateButton, is_bitmap_button=True)
GENBITMAPBUTTON = _make_button_class(buttons.GenBitmapButton, is_bitmap_button=True)
GENBITMAPTEXTBUTTON = _make_button_class(
    buttons.GenBitmapTextButton, is_bitmap_button=True
)
GENBITMAPBUTTONTXT = _make_button_class(BitmapTextButton, is_bitmap_button=True)
GENBITMAPBUTTONTXT_SMALL = _make_button_class(
    BitmapTextButton, is_bitmap_button=True, icon_size=1
)
NOBG_BUTTON = GENBITMAPBUTTONTXT
NOBG_BUTTON_TXT = GENBITMAPBUTTONTXT
CLOSEBUTTON = _make_button_class(
    BitmapTextButton, is_bitmap_button=True, is_close_button=True
)


def _make_menu_button_class(base_class):
    class _MENUBUTTON(base_class):
        def __init__(self, parent, **kwds):
            base_class.__init__(self, parent, **kwds)
            ldata = self.get_ldata()
            self._menu = None
            self.href_dict = {}
            if ldata:
                menu = wx.Menu()
                for row in ldata:
                    if row[0].replace("-", "").strip():
                        menu.Append(wx.NewId(), row[0], row[2]["href"])
                        self.href_dict[row[0]] = row[2]
                    else:
                        menu.AppendSeparator()
                self.SetMenu(menu)

            self.Bind(wx.EVT_MENU, self.on_menu)
            self.Bind(wx.EVT_BUTTON, self.on_button)

        def on_button(self, evt):
            self.ShowMenu()

        def on_menu(self, evt):
            e_obj = evt.GetEventObject()
            mitem = e_obj.FindItemById(evt.GetId())
            if mitem != wx.NOT_FOUND:
                label = mitem.GetItemLabel()
                item = self.href_dict[label]
                self.GetParent().href_clicked(self, item)

        def SetMenu(self, menu):
            self._menu = menu

        def ShowMenu(self):
            self.PopupMenu(self._menu, 1, self.GetSize()[1] - 1)

    return _MENUBUTTON


MENUBUTTON = _make_menu_button_class(SIMPLE_BUTTON)
MENUTOOLBARBUTTON = _make_menu_button_class(GENBITMAPBUTTONTXT)


class CHECKBOX(wx.CheckBox, SchBaseCtrl):
    """CHECKBOX handle ctrlcheckbox tag

    Tag arguments:
        checked - check checkbock
        value - if True checkbock is check else is not check
        label - checkbox label
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if "checked" in kwds:
            del kwds["checked"]
            checked = True
        else:
            checked = False

        if self.label:
            kwds["label"] = self.label

        if "value" in kwds:
            self.value = kwds["value"]
            del kwds["value"]
        else:
            self.value = None

        if not self.label:
            if "style" in kwds:
                kwds["style"] |= wx.ALIGN_RIGHT
            else:
                kwds["style"] = wx.ALIGN_RIGHT

        wx.CheckBox.__init__(self, parent, **kwds)
        if checked:
            self.SetValue(True)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        if "checked" in kwds:
            del kwds["checked"]
            checked = True
        else:
            checked = False
        if "value" in kwds:
            self.value = kwds["value"]
            del kwds["value"]
        else:
            self.value = None

        if checked:
            self.SetValue(True)

    def GetValue(self):
        value = wx.CheckBox.GetValue(self)
        if value == True:
            if self.value:
                return str(self.value)
            else:
                return True
        else:
            return None

    def SetValue(self, value):
        if type(value) == bool:
            return wx.CheckBox.SetValue(self, value)
        else:
            self.value = value


@handle_best_size
class CHECKLISTBOX(wx.CheckListBox, SchBaseCtrl):
    """CHECKBOXLIST handle ctrlchecklistbox tag

    Tag arguments:
        checked - check checkbock
        value - if True checkbock is check else is not check
        label - checkbox label
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)

        if "style" in kwds:
            kwds["style"] |= wx.WANTS_CHARS
        else:
            kwds["style"] = wx.WANTS_CHARS
        wx.CheckListBox.__init__(self, parent, **kwds)
        self.init_data(self.get_tdata())

    def init_data(self, data):
        checked_items = []
        if data:
            for row in data:
                id = self.Append(row[0].data)
                if "selected" in row[0].attrs:
                    checked_items.append(id)
        if checked_items:
            self.SetCheckedItems(checked_items)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        self.Clear()
        self.init_data(self.get_tdata())

    def GetBestSize(self):
        return (300, 68)

    def GetMinSize(self):
        return self.GetBestSize()

    def GetValue(self):
        ret = []
        data = self.get_tdata()
        items = self.GetCheckedItems()
        for item in items:
            if hasattr(data[item][0], "attrs") and "value" in data[item][0].attrs:
                ret.append(data[item][0].attrs["value"])
            else:
                ret.append(data[item][0].data)
        return ret


class BITMAPCOMBOBOX(BitmapComboBox, SchBaseCtrl):
    """BITMAPCOMBOBOX handle ctrlbitmapcombobox tag"""

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        BitmapComboBox.__init__(self, parent, **kwds)
        self.init_default_icons = False
        self._first_use = True
        self.accept_focus = False

    def after_create(self):
        super().after_create()
        if self.init_default_icons:

            def _fun():
                self.init_embeded_icons()
                self.init_fa_icons()

            wx.CallAfter(_fun)

    def init_wx_icons(self):
        for id in sorted([pos for pos in dir(wx) if pos.startswith("ART_")]):
            artid = getattr(wx, id)
            bmp = wx.ArtProvider.GetBitmap(artid, wx.ART_TOOLBAR, (22, 22))
            if bmp.IsOk():
                self.Append(id, bmp, id)

    def init_embeded_icons(self):
        base_path = wx.GetApp().src_path + "/static/icons/22x22/"
        return self._init_icons(base_path, "png://")

    def init_fa_icons(self):
        base_path = wx.GetApp().src_path + "/static/fonts/fork-awesome/fonts/22x22/"
        return self._init_icons(base_path, "fa://")

    def init_extern_icons(self, base_path, prefix):
        return self._init_icons(base_path, prefix)

    def _init_icons(self, base_path, prefix, subpath=None):
        if subpath:
            dirname = os.path.join(base_path, subpath)
        else:
            dirname = base_path

        if os.path.exists(dirname):
            for ff in os.listdir(dirname):
                if os.path.isdir(os.path.join(dirname, ff)):
                    if subpath:
                        self._init_icons(base_path, prefix, os.path.join(subpath, ff))
                    else:
                        self._init_icons(base_path, prefix, ff)
                else:
                    if ".png" in ff.lower():
                        try:
                            path = dirname + "/" + ff
                            image = wx.Image(path)
                            bmp = wx.Bitmap(image)
                            if subpath:
                                id = prefix + subpath + "/" + ff
                            else:
                                id = prefix + ff
                            self.Append(id.replace("\\", "/"), bmp, id)
                        except:
                            pass
                wx.Yield()

    def GetValue(self):
        ret = super().GetValue()
        if ret:
            return ret
        else:
            return ""

    def SetValue(self, value):
        if value:
            ret = super().SetValue(value)
        else:
            ret = super().SetValue("")
        return ret


class GAUGE(wx.Gauge, SchBaseCtrl):
    """GAUBE handle ctrlgauge tag

    Tag arguments:
        value - if True checkbock is check else is not check
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.Gauge.__init__(self, parent, **kwds)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        if self.value:
            self.SetValue(self.value)


class LISTBOX(wx.ListBox, SchBaseCtrl):
    """LISTBOX handle ctrllistbox tag

    Tag arguments:
        value - if True checkbock is check else is not check
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        self._norefresh = False

        if self.param and "multiple" in self.param:
            style = 0
            if "style" in kwds:
                style = kwds["style"]
            style = style | wx.LB_MULTIPLE
            kwds["style"] = style

        wx.ListBox.__init__(self, parent, **kwds)

        self.init_choices(self.get_tdata())

    def block_refresh(self):
        self._norefresh = True

    def find_and_select(self, item):
        """find first occurrence of item in list and select it"""
        l = self.GetStrings()
        for i in range(len(l)):
            if l[i].startswith(item):
                self.SetSelection(i)
                return

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        if not self._norefresh:
            self.Clear()
            self.refresh_tdata()
            self.init_choices(self.get_tdata())

    def init_choices(self, data):
        self.choices = []
        self.selected_rows = []
        if data:
            i = 0
            for row in data:
                value = row[0].data
                self.Append(value)
                if "value" in row[0].attrs:
                    self.choices.append(row[0].attrs["value"])
                else:
                    self.choices.append(value)
                if "selected" in row[0].attrs:
                    self.selected_rows.append(i)
                i += 1
        if self.selected_rows:
            for j in self.selected_rows:
                self.SetSelection(j)

    def GetValue(self):
        ret = []
        sel = self.GetSelections()
        for s in sel:
            ret.append(self.choices[s])
        return ret


class LIST(wx.ListCtrl, SchBaseCtrl):
    """LIST handle ctrllist tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)

        if "style" in kwds:
            kwds["style"] = kwds["style"] | wx.LC_REPORT
        else:
            kwds["style"] = wx.LC_REPORT
        kwds["size"] = wx.Size(-1, -1)

        wx.ListCtrl.__init__(self, parent, **kwds)

        tdata = self.get_tdata()
        l = 0
        if tdata:
            test = False
            for row in tdata:
                if test == False:
                    test = True
                    l = len(row)
                    for i in range(l):
                        self.InsertColumn(i, row[i].data)
                else:
                    index = self.InsertItem(0, row[0].data)
                    for i in range(1, l):
                        self.SetItem(index, i, row[i].data)

            for i in range(0, l):
                self.SetColumnWidth(i, -1)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        self.ClearAll()
        tdata = self.get_tdata()
        if tdata:
            test = False
            for row in tdata:
                if test == False:
                    test = True
                    l = len(row)
                    for i in range(l):
                        self.InsertColumn(i, row[i].data)
                else:
                    index = self.InsertStringItem(0, row[0].data)
                    for i in range(1, l):
                        self.SetStringItem(index, i, row[i].data)


class CHECKLIST(LIST):
    def __init__(self, parent, **kwds):
        super().__init__(parent, **kwds)
        self.EnableCheckBoxes()
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnItemActivated)

    def ToggleItem(self, index):
        toggle = not self.IsItemChecked(index)
        self.CheckItem(index, toggle)

    def OnItemActivated(self, evt):
        self.ToggleItem(evt.Index)

    def GetBestSize(self):
        return (400, 48)


class TABLE(SchGridPanel, SchBaseCtrl):
    """TABLE handle ctrltable tag

    Tag arguments:
        value
    """

    def get_table_lp(self):
        return self._table_lp

    def set_table_lp(self, table_lp):
        self._table_lp = table_lp

    table_lp = property(get_table_lp, set_table_lp)

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if self.param and "table_lp" in self.param:
            self.table_lp = int(self.param["table_lp"])
        else:
            self.table_lp = 0

        if "name" in kwds:
            name = kwds["name"]
        else:
            name = "LIST"

        if "size" in kwds:
            SchGridPanel.__init__(self, parent, size=kwds["size"], name=name)
        else:
            SchGridPanel.__init__(self, parent, name=name)

        tdata = self.get_tdata()
        if not tdata:
            print("no tdata:", self.href, self.src)
        if tdata:
            table = SimpleDataTable(self, tdata)
            if self.param and "no_actions" in self.param:
                table.set_no_actions(True)
        else:
            table = None

        self.grid = grid.SchTableGrid(
            table,
            "",
            self,
            typ=grid.SchTableGrid.VIEW,
            style=wx.TAB_TRAVERSAL | wx.FULL_REPAINT_ON_RESIZE,
        )
        # self.get_parent_page().register_signal(self, "refresh_controls")
        self.create_toolbar(self.grid)
        self.Bind(wx.EVT_CLOSE, self.on_close)
        self._table = table
        for signal_name in ("update_row_ok", "new_row_ok", "delete_row_ok"):
            self.get_parent_page().register_signal(self, signal_name)

    def on_close(self, event):
        for signal_name in ("update_row_ok", "new_row_ok", "delete_row_ok"):
            self.get_parent_page().unregister_signal(self, signal_name)
        event.Skip()

    def delete_row_ok(self, data):
        row_id = self.grid.GetGridCursorRow()
        self.grid.HideRow(row_id)
        row_id += 1
        if row_id >= self.grid.GetNumberRows():
            self.grid.goto_last_row()
        else:
            self.grid.SetGridCursor(row_id, 0)
            self.grid.MakeCellVisible(row_id, 0)

        return True

    def update_row_ok(self, data):
        return self._row_ok(data, insert=False)

    def new_row_ok(self, data):
        return self._row_ok(data, insert=True)

    def _row_ok(self, data, insert=False):
        url = self.GetParent().address
        if url:
            if "id" in data:
                pk = data["id"]
            else:
                pk = data["pk"]
            if "?" in url:
                url += "&pk=" + str(pk)
            else:
                url += "?pk=" + str(pk)
            http = wx.GetApp().get_http(self)
            response = http.get(self, url)
            if response.ret_code == 404:
                return
            data = response.str()
            tdatabuf = tdata_from_html(data, wx.GetApp().http)
            if len(tdatabuf) == 2:
                row = tdatabuf[1]
                if insert:
                    count = self.grid.GetNumberRows()
                    self.grid.GetTable().append_row(row)
                    self.grid.GetTable().refr_count(count + 1, False)
                    self.grid.GetTable().GetView().ForceRefresh()
                    self.grid.goto_last_row()
                else:
                    row_id = self.grid.GetGridCursorRow()
                    self.grid.GetTable().set_rec(row_id, row)
                    self.grid.GetTable().GetView().ForceRefresh()
                    self.grid.MakeCellVisible(self.grid.GetGridCursorRow(), 0)
                return True

    def GetMinSize(self):
        return SchGridPanel.GetMinSize(self)

    def process_refr_data(self, **kwds):
        self.grid.last_action = ""
        self.init_base(kwds)
        tdata = self.get_tdata()
        return self.do_refresh(tdata)

    def do_refresh(self, tdata):
        oldRow = self.grid.GetGridCursorRow()
        self._table.replace_tab(tdata)
        # self.grid.AutoSizeColumns(False)
        # self.grid.AutoSizeRows(True)
        if self.grid.last_action == "insert":
            newRow = self.grid.GetGridCursorRow() + 1
            if newRow < self.grid.GetTable().GetNumberRows():
                self.grid.SetGridCursor(newRow, 0)
                self.grid.MakeCellVisible(newRow, 0)
        else:  # self.grid.last_action == 'edit':
            if oldRow < self.grid.GetTable().GetNumberRows():
                if oldRow < 0:
                    oldRow = 0
                self.grid.SetGridCursor(oldRow, 0)
                self.grid.MakeCellVisible(oldRow, 0)
            else:
                self.grid.goto_last_row()

    def refresh_from_source(self, html_src):
        self.refresh_tdata(html_src)
        tdata = self.get_tdata()
        return self.do_refresh(tdata)

    # def refresh_controls(self):
    #    self.get_parent_form().enable_ctrls((self,))
    #    ret = self.get_parent_page()._refresh_html()
    #    self.get_parent_form().enable_ctrls(None)
    #    return ret


class RADIOBOX(wx.RadioBox, SchBaseCtrl):
    """RADIOBOX handle ctrlradiobox tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        tdata = self.get_tdata()
        if tdata:
            choices = []

            for row in tdata:
                choices.append(row[0].data)

            kwds["choices"] = choices
        wx.RadioBox.__init__(self, parent, **kwds)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        self.Clear()
        self.refresh_tdata()
        tdata = self.get_tdata()
        if tdata:
            for row in tdata:
                self.Append(row[0].data)


class RADIOBUTTON(wx.RadioButton, SchBaseCtrl):
    """RADIOBUTTON handle ctrlradiobutton tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if "name" in kwds:
            name = kwds["name"]
            if not hasattr(parent, name.split("__")[0]):
                if "style" in kwds:
                    kwds["style"] |= wx.RB_GROUP
                else:
                    kwds["style"] = wx.RB_GROUP

        if "checked" in kwds:
            del kwds["checked"]
            checked = True
        else:
            checked = False
        if "value" in kwds:
            self.value = kwds["value"]
            del kwds["value"]
        else:
            self.value = None

        wx.RadioButton.__init__(self, parent, **kwds)
        if checked:
            self.SetValue(True)

    def GetValue(self):
        value = wx.RadioButton.GetValue(self)
        if value == True:
            if self.value:
                return str(self.value)
            else:
                return None
        else:
            return None

    def SetValue(self, value):
        if value.__class__ == bool:
            return wx.RadioButton.SetValue(self, value)
        else:
            self.value = value


class SLIDER(wx.Slider, SchBaseCtrl):
    """SLIDER handle ctrlslider tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.Slider.__init__(self, parent, **kwds)


# class SPINBUTTON(wx.SpinButton, SchBaseCtrl):
#    """SPINBUTTON handle ctrlspinbutton tag
#
#    Tag arguments:
#        value
#    """
#
#    def __init__(self, parent, **kwds):
#        SchBaseCtrl.__init__(self, parent, kwds)
#        wx.SpinButton.__init__(self, parent, **kwds)


class SPIN(wx.SpinCtrl, SchBaseCtrl):
    """SPIN handle ctrlspinbutton tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.SpinCtrl.__init__(self, parent, **kwds)


class STATICTEXT(wx.StaticText, SchBaseCtrl):
    """STATICTEXT handle ctrlstatictext tag

    Tag arguments:
        label
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.StaticText.__init__(self, parent, **kwds)
        if self.label:
            self.SetLabel(self.label)


class ERRORLIST(BitmapTextButton, SchBaseCtrl):
    """ERRORLIST handle ctrlerrorlist tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)

        self.bmp = bitmap_from_href("client://status/dialog-error.png", 1)

        kwds["style"] = wx.NO_BORDER
        kwds["bitmap"] = self.bmp

        BitmapTextButton.__init__(self, parent, **kwds)
        if self.ldata:
            self.SetToolTip(wx.ToolTip(self.ldata[0][0]))
        else:
            if "data" in self.param:
                self.SetToolTip(wx.ToolTip(self.param["data"]))


class TEXT(SchBaseCtrl, wx.TextCtrl):
    """TEXT handle ctrltext tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if self.param and "process_enter" in self.param:
            kwds["style"] = wx.TE_PROCESS_ENTER
        wx.TextCtrl.__init__(self, parent, **kwds)
        if self.hidden:
            self.Enable(False)
        if self.maxlength:
            self.SetMaxLength(int(self.maxlength))

    def SetValue(self, value):
        if type(value) == str:
            return wx.TextCtrl.SetValue(self, value)
        else:
            return wx.TextCtrl.SetValue(self, value.decode("utf-8"))

    def GetValue(self):
        return super().GetValue()

    def GetBestSize(self):
        size = super().GetBestSize()
        return (4 * size[0], size[1])


class SEARCH(wx.SearchCtrl, SchBaseCtrl):
    """SEARCH handle ctrlsearch tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if wx.Platform in [
            "__WXGTK__",
        ]:
            kwds["size"] = (200, 28)
        else:
            kwds["size"] = (200, -1)
        if self.param and "process_enter" in self.param:
            kwds["style"] = wx.TE_PROCESS_ENTER

        wx.SearchCtrl.__init__(self, parent, **kwds)

        if wx.Platform in ["__WXGTK__", "__WXMSW__"]:
            for child in list(self.GetChildren()):
                if isinstance(child, wx.TextCtrl):
                    child.Bind(wx.EVT_KEY_DOWN, self.on_key_down_base)
                    break


class PASSWORD(TEXT, SchBaseCtrl):
    """PASSWORD handle ctrlpassword tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        if "style" in kwds:
            kwds["style"] = kwds["style"] | wx.TE_PASSWORD
        else:
            kwds["style"] = wx.TE_PASSWORD

        TEXT.__init__(self, parent, **kwds)


class TREE(wx.TreeCtrl, SchBaseCtrl):
    """TREE handle ctrltree tag

    Tag arguments:
        value
    """

    def _append_list(self, parent, list):
        for row in list:
            child = self.AppendItem(parent, row[0])
            self.SetItemData(child, dict(row[2]))
            if len(row[1]) > 0:
                self.SetItemImage(child, self.fldridx, wx.TreeItemIcon_Normal)
                self.SetItemImage(child, self.fldropenidx, wx.TreeItemIcon_Expanded)
                self._append_list(child, row[1])
            else:
                self.SetItemImage(child, self.fileidx, wx.TreeItemIcon_Normal)
                self.SetItemImage(child, self.fileidxmark, wx.TreeItemIcon_Selected)

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        kwds["style"] = wx.TR_HIDE_ROOT | wx.TR_DEFAULT_STYLE
        wx.TreeCtrl.__init__(self, parent, **kwds)

        isz = (16, 16)
        il = wx.ImageList(isz[0], isz[1])
        self.fldridx = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_FOLDER, wx.ART_OTHER, isz)
        )
        self.fldropenidx = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_FILE_OPEN, wx.ART_OTHER, isz)
        )
        self.fileidx = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_NORMAL_FILE, wx.ART_OTHER, isz)
        )
        self.fileidxmark = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_TICK_MARK, wx.ART_OTHER, isz)
        )

        self.SetImageList(il)
        self.il = il

        self.root = self.AddRoot("/")

        ldata = self.get_ldata()
        if ldata:
            self._append_list(self.root, ldata)

        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.on_activated)

    def on_activated(self, event):
        item_id = event.GetItem()
        if item_id.IsOk():
            item = self.GetPyData(item_id)
            # print "OnActivated:", item
            if "href" in item:
                self.GetParent().href_clicked(self, item)
        event.Skip()


class COLOURSELECT(colourselect.ColourSelect, SchBaseCtrl):
    """COLOURSELECT handle ctrlcolourselect tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if "name" in kwds:
            del kwds["name"]
        colourselect.ColourSelect.__init__(self, parent, id=-1, **kwds)


class GENERICDIR(wx.GenericDirCtrl, SchBaseCtrl):
    """GENERICDIR handle ctrlgenericdir tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.GenericDirCtrl.__init__(self, parent, **kwds)


class TREELIST(TreeListCtrl, SchBaseCtrl):
    """TREELIST handle ctrltreelist tag

    Tag arguments:
        value
    """

    def append_list(self, parent, list):
        for row in list:
            l = row[0].split("||")
            child = self.AppendItem(parent, l[0])
            try:
                self.SetItemData(child, None)
            except:
                pass

            for i in range(len(l) - 1):
                self.SetItemText(child, l[i + 1], i + 1)

            if len(row[1]) > 0:
                self.SetItemImage(child, self.fldridx, wx.TreeItemIcon_Normal)

                self.append_list(child, row[1])
            else:
                self.SetItemImage(child, self.fileidx, wx.TreeItemIcon_Normal)

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        kwds["agwStyle"] = wx.TR_HIDE_ROOT
        TreeListCtrl.__init__(self, parent, **kwds)
        isz = (16, 16)
        il = wx.ImageList(isz[0], isz[1])
        self.fldridx = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_FOLDER, wx.ART_OTHER, isz)
        )
        self.fldropenidx = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_FILE_OPEN, wx.ART_OTHER, isz)
        )
        self.fileidx = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_NORMAL_FILE, wx.ART_OTHER, isz)
        )
        self.fileidxmark = il.Add(
            wx.ArtProvider.GetBitmap(wx.ART_TICK_MARK, wx.ART_OTHER, isz)
        )

        self.SetImageList(il)
        self.il = il

        l = self.label.split("||")
        for pos in l:
            try:
                self.AppendColumn(pos)
            except:
                self.AddColumn(pos)

        self.SetColumnWidth(0, 175)
        self.root = self.AddRoot("/")
        ldata = self.get_ldata()
        if ldata:
            self.append_list(self.root, ldata)
        self.Refresh()


class CALENDAR(CalendarCtrl, SchBaseCtrl):
    """CALENDAR handle ctrlcalendar tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        try:
            CalendarCtrl.__init__(
                self,
                parent,
                style=wx.adv.CAL_MONDAY_FIRST
                | wx.adv.CAL_SHOW_HOLIDAYS
                | wx.adv.CAL_SEQUENTIAL_MONTH_SELECTION,
                **kwds,
            )
        except:
            CalendarCtrl.__init__(
                self,
                parent,
                style=wx.calendar.CAL_MONDAY_FIRST
                | wx.calendar.CAL_SHOW_HOLIDAYS
                | wx.calendar.CAL_SEQUENTIAL_MONTH_SELECTION,
                **kwds,
            )


class EDITABLELISTBOX(EditableListBox, SchBaseCtrl):
    """EDITABLELISTBOX handle ctrleditablelistbox tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        EditableListBox.__init__(self, parent, **kwds)


class FILEBROWSEBUTTON(filebrowsebutton.FileBrowseButton, SchBaseCtrl):
    """FILEBROWSEBUTTON handle ctrlfilebrowsebutton tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        kwds["labelText"] = ""
        kwds["buttonText"] = str(_("Browse"))
        kwds["size"] = (400, -1)
        filebrowsebutton.FileBrowseButton.__init__(self, parent, **kwds)

    def GetValue(self):
        return "@" + super(FILEBROWSEBUTTON, self).GetValue()


class IMAGEBROWSEBUTTON(FILEBROWSEBUTTON, SchBaseCtrl):
    """IMAGEBROWSEBUTTON handle ctrlimagebrowsebutton tag

    Tag arguments:
        value
    """

    def OnBrowse(self, event=None):
        current = self.GetValue()
        directory = os.path.split(current)
        if os.path.isdir(current):
            directory = current
            current = ""
        elif directory and os.path.isdir(directory[0]):
            current = directory[1]
            directory = directory[0]
        else:
            directory = self.startDirectory
            current = ""

        dlg = wx.lib.imagebrowser.ImageDialog(self, current)

        if dlg.ShowModal() == wx.ID_OK:
            self.SetValue(dlg.GetFile())
        dlg.Destroy()


class MASKTEXT(masked.TextCtrl, SchBaseCtrl):
    """MASKTEXT handle ctrlmasktext tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if self.valuetype:
            if self.src:
                if self.src.startswith("!"):
                    kwds["autoformat"] = self.src[1:]
                else:
                    kwds["mask"] = self.src
        masked.TextCtrl.__init__(self, parent, **kwds)
        self._autofit = False


class NUM(wx.SpinCtrl, SchBaseCtrl):
    """NUM handle ctrlnum tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if self.param and "process_enter" in self.param:
            kwds["style"] = wx.TE_PROCESS_ENTER
        if self.readonly:
            style = 0
            if "style" in kwds:
                style = kwds["style"]
            style = style | wx.TE_READONLY
            kwds["style"] = style
        wx.SpinCtrl.__init__(self, parent, **kwds)


class AMOUNT(wx.SpinCtrlDouble, SchBaseCtrl):
    """handle ctrlfloat tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if self.param and "process_enter" in self.param:
            kwds["style"] = wx.TE_PROCESS_ENTER
        if self.readonly:
            style = 0
            if "style" in kwds:
                style = kwds["style"]
            style = style | wx.TE_READONLY
            kwds["style"] = style

        kwds["inc"] = 1

        if self.param and "min" in self.param:
            kwds["min"] = float(self.param["min"])
        else:
            kwds["min"] = -100000000
        if self.param and "max" in self.param:
            kwds["max"] = float(self.param["max"])
        else:
            kwds["max"] = 100000000

        kwds["style"] = wx.SP_ARROW_KEYS | wx.ALIGN_RIGHT

        wx.SpinCtrlDouble.__init__(self, parent, **kwds)
        self.SetDigits(2)


class FLOAT(wx.SpinCtrlDouble, SchBaseCtrl):
    """handle ctrlfloat tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if self.param and "process_enter" in self.param:
            kwds["style"] = wx.TE_PROCESS_ENTER
        if self.readonly:
            style = 0
            if "style" in kwds:
                style = kwds["style"]
            style = style | wx.TE_READONLY
            kwds["style"] = style

        kwds["inc"] = 1

        if self.param and "min" in self.param:
            kwds["min"] = float(self.param["min"])
        else:
            kwds["min"] = -100000000
        if self.param and "max" in self.param:
            kwds["max"] = float(self.param["max"])
        else:
            kwds["max"] = 100000000

        kwds["style"] = wx.SP_ARROW_KEYS | wx.ALIGN_RIGHT

        wx.SpinCtrlDouble.__init__(self, parent, **kwds)

        self.SetDigits(6)


class TIME(masked.TimeCtrl, SchBaseCtrl):
    """TIME handle ctrltime tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        masked.TimeCtrl.__init__(self, parent, **kwds)

        # self.SetEmptyBackgroundColour(wx.Colour(1,0,0))
        # self.SetValidBackgroundColour("Red")
        # self.SetInvalidBackgroundColour(wx.Colour(1,0,0))
        # self._validBackgroundColour = wx.Colour(1,0,0)


class STYLEDTEXT(wx.TextCtrl, SchBaseCtrl):
    """STYLEDTEXT handle ctrlstyledtext tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        if "style" in kwds:
            style = kwds["style"]
            style = style | wx.TE_MULTILINE
            kwds["style"] = style
        else:
            kwds["style"] = wx.TE_MULTILINE
        wx.TextCtrl.__init__(self, parent, **kwds)
        if "data" in self.param:
            self.SetValue(self.param["data"])

    def SetValue(self, value):
        if value.startswith("\n"):
            value2 = value[1:]
        else:
            value2 = value
        if type(value2) == str:
            return wx.TextCtrl.SetValue(self, value2)
        else:
            return wx.TextCtrl.SetValue(self, value2.decode("utf-8"))


AUTOCOMPLETE = STYLEDTEXT
STANDARDSTYLEDTEXT = STYLEDTEXT


def TEXTAREA(parent, **kwds):
    """TEXTAREA handle ctrltextarea tag

    Tag arguments:
        value
    """

    if "param" in kwds:
        if "data" in kwds["param"]:
            data = kwds["param"]["data"].replace("\r", "")
            if data.startswith("\n"):
                data = data[1:]
                kwds["param"]["data"] = data

    size = [-1, -1]

    if not "width" in kwds["param"]:
        if "cols" in kwds["param"]:
            width = 7 * int(is_null(kwds["param"]["cols"], "80"))
            if width > 480:
                width = 480
            kwds["param"]["width"] = str(width)
            size[0] = width
        else:
            kwds["param"]["width"] = "100%"
    if not "height" in kwds["param"]:
        if "rows" in kwds["param"]:
            size[1] = 14 + 14 * int(is_null(kwds["param"]["rows"], "3"))
            kwds["param"]["height"] = str(size[1])
        else:
            kwds["param"]["height"] = "100%"

    kwds["size"] = size

    if "cols" in kwds:
        del kwds["cols"]

    if "rows" in kwds:
        del kwds["rows"]

    if "src" in kwds:
        if kwds["src"] in ("c", "python", "html"):
            ret = STYLEDTEXT(parent, **kwds)
        else:
            ret = AUTOCOMPLETE(parent, **kwds)
    else:
        ret = STYLEDTEXT(parent, **kwds)
    return ret


class TICKER(ticker.Ticker, SchBaseCtrl):
    """TICKER handle ctrlticker tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        ticker.Ticker.__init__(self, parent, **kwds)

    def SetValue(self, value):
        self.SetText(value)

    def CanClose(self):
        self.Stop()
        return True


INIT_CSS_STR = """
body {font-family:sans-serif;font-size:150%; padding:1;}
table {border:0;vertical-align:top; padding:1;}
td table { padding: 1; }
th {border:0; cellpadding:1;}
td {border:0; vertical-align:top; cellpadding:1;}
strong,b {font-weight:bold;}
p { cellpadding:1; border:0; width:100%; }
"""


class HTMLLISTBOX(wx.VListBox, SchBaseCtrl):
    """HTMLLISTBOX handle ctrlhtmllist tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        self.choices = []
        self.getItemFunct = None
        self.h = None
        if self.param and "multiple" in self.param:
            style = 0
            if "style" in kwds:
                style = kwds["style"]
            style = style | wx.LB_MULTIPLE
            kwds["style"] = style

        if "size" in kwds:
            size = kwds["size"]
            if size[1] == -1:
                kwds["size"] = (size[0], 100)

        tdata = self.get_tdata()
        if tdata:
            for row in tdata:
                if hasattr(row[0], "attrs") and "value" in row[0].attrs:
                    if "selected" in row[0].attrs:
                        selected = True
                    else:
                        selected = False
                    self._append_html(row[0].attrs["value"], selected, row[0].data)
                else:
                    self.append_html(row[0].data)

        wx.VListBox.__init__(self, parent, **kwds)

        self.SetItemCount(len(self.choices))
        i = 0
        for choice in self.choices:
            if choice[1]:
                self.SetSelection(i)
            i += 1

        self.ScrollToRow(0)

    def _append_html(self, id, sel, html_txt):
        self.choices.append((id, sel, html_txt.replace("[", "<").replace("]", ">")))

    def append_html(self, html_txt):
        self._append_html(-1, False, html_txt)

    def append_text(self, txt):
        self.append_html(txt.replace("\n", ""))
        self.SetItemCount(len(self.choices))
        self.Refresh()

    def append_texts(self, txt_list):
        for txt in txt_list:
            self.append_html(txt.replace("\n", ""))
        self.SetItemCount(len(self.choices))
        self.SetSelection(len(self.choices) - 1)
        self.Refresh()

    def GetValue(self):
        ret = []
        if self.HasMultipleSelection():
            item, cookie = self.GetFirstSelected()
            while item != wx.NOT_FOUND:
                ret.append(self.choices[item][0])
                item, cookie = self.GetNextSelected(cookie)
        return ret

    def _calc_or_draw(self, n, dc, rect, calc_only):
        w = self.GetSize()[0]
        value = "<html><body>" + (self.choices)[n][2] + "</body></html>"
        wxdc = DcDc(dc, calc_only=calc_only, width=w, height=-1)

        if rect:
            wxdc2 = wxdc.subdc(rect.X + 2, rect.Y, rect.Width - 4, rect.Height)
        else:
            wxdc2 = wxdc

        p = HtmlViewerParser(dc=wxdc2, calc_only=calc_only, init_css_str=INIT_CSS_STR)
        p.set_http_object(wx.GetApp().http)
        p.set_parent_window(self)
        try:
            p.feed(value)
        except:
            print("ERROR:", value)

        w2, h2 = p.get_max_sizes()
        if not self.h:
            self.h = h2
        return w2, h2

    def OnMeasureItem(self, n):
        dc = wx.ClientDC(self)
        w, h = self._calc_or_draw(n, dc, None, True)
        return h

    def OnDrawItem(self, dc, rect, n):
        self._calc_or_draw(n, dc, rect, False)

    def CanAcceptFocus(self):
        return False

    def GetBestSize(self):
        if self.h:
            return (400, self.h + 2)
        else:
            return (400, 200)

    def scroll_to_line(self, line_no):
        nr = line_no
        count = self.GetItemCount()
        if count > 0:
            if nr >= count:
                nr = count - 1
            if nr < 0:
                nr = 0
            self.SetSelection(nr)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)
        self.choices = []
        tdata = self.get_tdata()

        if tdata:
            for row in tdata:
                if hasattr(row[0], "attrs") and "value" in row[0].attrs:
                    if "selected" in row[0].attrs:
                        selected = True
                    else:
                        selected = False
                    self._append_html(row[0].attrs["value"], selected, row[0].data)
                else:
                    self.append_html(row[0].data)
        self.SetItemCount(len(self.choices))
        i = 0
        for choice in self.choices:
            if choice[1]:
                self.SetSelection(i)
            i += 1

        self.ScrollToRow(0)

        # if tdata:
        #    for row in tdata:
        #        self.append_html(row[0].data)


#        self.SetItemCount(len(self.choices))
#        self.SetSelection(len(self.choices) - 1)


class HTML(page.SchPage, SchBaseCtrl):
    """HTML handle ctrlhtml tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        page.SchPage.__init__(self, parent, self.src, None)
        if self.src:
            value = self.load_string_from_server(self.src)
            self._set_value(value, False)
        elif self.param and "data" in self.param and len(self.param["data"]) > 0:
            self._set_value(self.param["data"], False)
        else:
            self._value = None
        self.body.Refresh()
        self.body.Update()

    def CanAcceptFocus(self):
        return False

    def SetValue(self, value):
        self._set_value(value)

    def _set_value(self, value, refresh=True):
        if not "<html" in value:
            self._value = "<html><body>%s</body></html>" % value
        else:
            self._value = value
        mp = ShtmlParser()
        mp.process(self._value)
        mp.address = None
        body = mp.get_body()
        self.body.show_page(body)
        if refresh:
            self.body.wxdc = None
            self.body.draw_background()

    def GetValue(self):
        return self._value


HTML2 = None


class NOTEBOOK(wx.Notebook, SchBaseCtrl):
    """NOTEBOOK handle ctrlnotebook tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        self.children = []
        wx.Notebook.__init__(self, parent, **kwds)
        if self.tdata:
            for row in self.tdata:
                h = page.SchPage(self, row[1].data, {})
                self.AddPage(h, row[0].data)
                self.children.append(h)


class GRID(grid.SchTableGrid, SchBaseCtrl):
    """GRID handle ctrlgrid tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        parm = createparm.create_parm(self.src, parent.get_parm_obj())
        if parm:
            self.proxy = tabproxy.DataProxy(wx.GetApp().get_http(parent), str(parm[0]))
            self.proxy.set_address_parm(parm[2])
        else:
            self.proxy = tabproxy.DataProxy(wx.GetApp().get_http(parent), str(self.src))
        table = gridtable_from_proxy.DataSource(self.proxy)

        if self.readonly:
            kwds["typ"] = self.VIEW
            table.set_read_only(True)
        else:
            table.set_read_only(False)

        super().__init__(table, self.src, parent, **kwds)

    def process_refr_data(self, **kwds):
        self.init_base(kwds)

        parm = createparm.create_parm(self.src, self.GetParent().get_parm_obj())
        if parm:
            self.proxy = tabproxy.DataProxy(wx.GetApp().get_http(self), str(parm[0]))
            self.proxy.SetAddressParm(parm[2])
        else:
            self.proxy = tabproxy.DataProxy(wx.GetApp().get_http(self), str(self.src))
        table = gridtable_from_proxy.DataSource(self.proxy)
        self.SetTable(table)

    def refr_obj(self):
        if self.GetParent().IsShown():
            parm = createparm.create_parm(self.src, self.GetParent().get_parm_obj())
            if parm:
                if self.proxy.set_address_parm(parm[2]):
                    self.GetTable().refresh(False)

    def OnSize(self, event=None):
        """Handles The wx.EVT_SIZE Event For TabCtrl."""

        if event != None:
            size = event.GetSize()
            old = self.GetSize()
            if old[1] != size[1] - 8:
                self.SetSize((size[0], size[1] - 8))
                self.GetParent().RefrPage()
            event.Skip()


class UPDATEGRIDBUTTON(wx.Button, SchBaseCtrl):
    """UPDATEGRIDBUTTON handle ctrlupdategridbutton tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.Button.__init__(self, parent, **kwds)

        self.Bind(wx.EVT_BUTTON, self.OnClick, self)
        self.Enable()

    def OnClick(self, event):
        rec = dict()
        keys = list(self.GetParent().GetWidgets().keys())
        for k in keys:
            ctrl = self.GetParent().GetItem(k)
            if hasattr(ctrl, "GetValue"):
                rec[k] = self.GetParent().GetItem(k).GetValue()
        self.GetParent().GetPrevWin().ActiveCtrl.OnUpRecFromForm(rec)
        self.GetParent().any_parent_command("OnCancel", None)


class POPUPHTML(DataPopupControl, SchBaseCtrl):
    """POPUPHTML handle ctrlpopuphtml tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        """Constructor

        href - base address,
            href + "dialog" - address of dialog window
            href+ "test?value=" - address of records which starts with value parameter
        """

        SchBaseCtrl.__init__(self, parent, kwds)
        DataPopupControl.__init__(self, parent, **kwds)
        if self.param:
            if "IN_NEW_WIN" in self.param:
                self.simpleDialog = False

    def GetBestSize(self):
        dx, dy = DataPopupControl.GetBestSize(self)
        return (250, dy)


if platform.system() == "Linux":

    class DATEPICKER(POPUPHTML):
        """DATEPICKER handle ctrldatepicker tag

        Tag arguments:
            value
        """

        def __init__(self, parent, **kwds):
            kwds["href"] = wx.GetApp().make_href("/schsys/datedialog/")

            if "style" in kwds:
                kwds["style"] = kwds["style"] | wx.WANTS_CHARS
            else:
                kwds["style"] = wx.WANTS_CHARS

            if "size" in kwds:
                kwds["size"] = wx.Size(150, kwds["size"].height)
            else:
                kwds["size"] = wx.Size(150, -1)

            POPUPHTML.__init__(self, parent, **kwds)

            if self.value:
                self.set_rec(self.value, Td(self.value))
            else:
                self.set_rec(
                    wx.DateTime.Today().FormatISODate(),
                    Td(wx.DateTime.Today().FormatISODate()),
                    False,
                )

            wx.CallAfter(
                self.to_masked,
                mask="####-##-##",
                formatcodes="F",
                validRegex=r"\d{4}-\d{2}-\d{2}",
            )
            # wx.CallAfter(self.to_masked,autoformat="EUDATEYYYYMMDD.")

        def GetValue(self):
            value = self.get_rec()
            # if type(value) == tuple and len(value)>0:
            #    return value[1][0]
            if value.data:
                return value.data
            else:
                return value
            return None

        def GetBestSize(self):
            dx, dy = POPUPHTML.GetBestSize(self)
            return (130, dy)

else:

    class DATEPICKER(DatePickerCtrl, SchBaseCtrl):
        def __init__(self, parent, **kwds):
            SchBaseCtrl.__init__(self, parent, kwds)
            kwds["size"] = (120, -1)
            try:
                kwds["style"] = (
                    wx.adv.DP_DROPDOWN | wx.adv.DP_SHOWCENTURY | wx.adv.DP_ALLOWNONE
                )
            except:
                kwds["style"] = wx.DP_DROPDOWN | wx.DP_SHOWCENTURY | wx.DP_ALLOWNONE
            DatePickerCtrl.__init__(self, parent, **kwds)

        def SetValue(self, value):
            if value:
                date = wx.DateTime()
                date.ParseDate(value)
                DatePickerCtrl.SetValue(self, date)
            else:
                DatePickerCtrl.SetValue(self, None)

        def GetValue(self):
            value = DatePickerCtrl.GetValue(self)
            if value:
                return value.FormatISODate()
            else:
                return None


class DATETIMEPICKER(POPUPHTML):
    """DATETIMEPICKER handle ctrldatetimepicker tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        kwds["href"] = wx.GetApp().make_href("/schsys/datedialog/")

        if "style" in kwds:
            kwds["style"] = kwds["style"] | wx.WANTS_CHARS
        else:
            kwds["style"] = wx.WANTS_CHARS

        if "size" in kwds:
            kwds["size"] = wx.Size(200, kwds["size"].height)
        else:
            kwds["size"] = wx.Size(200, -1)

        POPUPHTML.__init__(self, parent, **kwds)

        # self.to_masked(autoformat="EUDATE24HRTIMEYYYYMMDD.HHMM")
        # self.to_masked(autoformat="EUDATE24HRTIMEYYYYMMDD.HHMM")
        self.to_masked(
            mask="####-##-## ##:##",
            formatcodes="F",
            validRegex=r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}",
        )
        # self.win.SetValidBackgroundColour("Red")
        # self.win.SetValidBackgroundColour(wx.Colour(1,0,0))
        # self.win.SetValidBackgroundColour(
        #    wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        # )

        if self.value:
            self.set_rec(self.value, Td(self.value))
        else:
            now = datetime.datetime.now().isoformat().replace("T", " ")[:16]
            self.set_rec(now, Td(now), False)

    def set_rec(self, value, value_rec, dismiss=False):
        if len(value) == 16:
            return super().set_rec(value, value_rec, dismiss)
        elif len(value) == 10:
            value_rec.data = value_rec.data + " 00:00"
            return super().set_rec(value + " 00:00", value_rec, dismiss)
        else:
            return super().set_rec("0000.00.00 00:00", value_rec, dismiss)

    def GetValue(self):
        value = self.get_rec()
        value2 = self.GetTextCtrl().GetValue()
        if value2 and value2[0] != " ":
            if value.data:
                return value.data
            else:
                return [
                    datetime.datetime.strptime(value2, "%Y-%m-%d %H:%M"),
                ]
        return None

    def GetBestSize(self):
        dx, dy = POPUPHTML.GetBestSize(self)
        return (180, dy)


class CHOICE(POPUPHTML):
    """CHOICE handle ctrlchoice tag

    Tag arguments:d
        value
    """

    def __init__(self, parent, **kwds):
        if not ("href" in kwds or "href" in kwds):
            kwds["href"] = wx.GetApp().make_href("/schsys/listdialog/")

        if "style" in kwds:
            kwds["style"] = kwds["style"] | wx.WANTS_CHARS
        else:
            kwds["style"] = wx.WANTS_CHARS

        if "size" not in kwds:
            kwds["size"] = wx.Size(250, -1)
        self.height = 250
        kwds["dialog_with_value"] = False
        POPUPHTML.__init__(self, parent, **kwds)

        self.init_choices(self.get_tdata())

        aTable = [
            (0, wx.WXK_F2, self.on_ext_button_click),
        ]
        self.set_acc_key_tab(aTable)

    def init_choices(self, data):
        self.choices = []
        if data:
            i = 0
            for row in data:
                self.choices.append(row[0])
                if "selected" in row[0].attrs:
                    value = row[0].data
                    ComboCtrl.SetValue(self, value)
                    if "value" in row[0].attrs:
                        self.set_rec(row[0].attrs["value"], row[0], dismiss=False)
                    else:
                        self.set_rec(value, row[0], dismiss=False)

    def on_ext_button_click(self, event):
        ret = self.alternate_button_click()
        if self.page:
            self.page.body.choices = self.choices
            wx.CallAfter(self.page.body.refr)
        else:
            self.popup.html.body.choices = self.choices
            wx.CallAfter(self.page.body.refr)
        return ret

    def OnButtonClick(self):
        if self.simpleDialog:
            ret = POPUPHTML.OnButtonClick(self)
            self.popup.html.body.choices = self.choices
            wx.CallAfter(self.popup.html.body.refr)
        else:
            ret = POPUPHTML.OnButtonClick(self)
            self.page.body.choices = self.choices
            wx.CallAfter(self.popup.html.body.refr)
        return ret

    def GetValue(self):
        if self.readonly:
            value = self.get_rec()
            return value[0]
        else:
            return POPUPHTML.GetValue(self)
        return None

    def GetBestSize(self):
        dx, dy = POPUPHTML.GetBestSize(self)
        return (dx, dy)

    def GetMinSize(self):
        return self.GetBestSize()

    def GetMaxSize(self):
        return self.GetBestSize()


class DBCHOICE(CHOICE):
    """DBCHOICE handle ctrldbchoice tag

    Tag arguments:
        value
    """

    def GetValue(self):
        if self.readonly:
            value = self.get_rec()
            if value:
                if hasattr(value, "attrs") and "value" in value.attrs:
                    return value.attrs["value"]
                else:
                    return value.data
            else:
                return None
        else:
            return POPUPHTML.GetValue(self)
        return None


class DBCHOICE_EXT(POPUPHTML):
    """DBCHOICE_EXT handle ctrldbchoice_ext tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        if "style" in kwds:
            kwds["style"] = kwds["style"] | wx.CB_READONLY
        else:
            kwds["style"] = wx.CB_READONLY

        POPUPHTML.__init__(self, parent, **kwds)

    def SetValue(self, value):
        if type(value) == dict:
            val = value["value"]
            title = value["title"]
            sel = value["selected"]
            if sel:
                self.set_rec(val, Td(title, {"value": val}), dismiss=False)
        else:
            if "!!" in value:
                id = value.split(":")[0]
                name = value[len(id) + 1 :].replace("!!", "")
                self.set_rec(name, [id, name], dismiss=False)

        POPUPHTML.SetValue(self, value)

    def GetValue(self):
        if self.readonly:
            value = self.get_rec()
            if value.__class__ == tuple and len(value) > 0:
                return value[1][0]
            if len(value) > 0:
                return value[0]
            else:
                return None
        else:
            return POPUPHTML.GetValue(self)
        return None

    def get_parm(self, parm):
        if len(self.rec_value) > 0:
            id = self.rec_value[0]
            return b64encode(str(id).encode("utf-8")) if parm == "value" else None
        else:
            return None


class COLLAPSIBLE_PANEL(wx.CollapsiblePane, SchBaseCtrl):
    """COLLAPSIBLE_PANEL handle ctrlcollapsible_panel tag

    Tag arguments:
        value
    """

    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        kwds["label"] = self.label
        kwds["style"] = wx.CP_DEFAULT_STYLE | wx.CP_NO_TLW_RESIZE
        if "size" in kwds:
            tmp = kwds["size"]
            self._size = [tmp[0], tmp[1]]
            del kwds["size"]
        else:
            self._size = [400, 300]

        self._height = 0

        if "collapse_height" in self.param:
            self._size[1] = int(self.param["collapse_height"])

        wx.CollapsiblePane.__init__(self, parent, **kwds)
        try:
            self.GetPane().SetBackgroundStyle(wx.BG_STYLE_ERASE)
        except:
            self.GetPane().SetBackgroundStyle(wx.BG_STYLE_CUSTOM)

        if "data" in self.param and self.param["data"].strip():
            self.set_html(self.param["data"])
        else:
            self.html = None
            self.SetSize((0, 0))
            self.Hide()
        self.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, self.on_pane_changed)

    def set_html(self, html_txt):
        mp = ShtmlParser()
        mp.process(
            "<html><body>" + self.param["data"].decode("utf-8") + "</body></html>"
        )
        mp.address = None
        if not self.html:
            self.html = page.SchPage(self.GetPane(), mp, {})
            self.html.init_frame()
            self.html.activate_page()
            width = self._size[0] if self._size[0] > 0 else 400
            self._height = self._size[1] if self._size[1] > 0 else 300
            self.html.SetSize((width, self._height))
            sizer = wx.BoxSizer()
            sizer.Add(self.html, 1, wx.EXPAND | wx.ALL, 5)
            self.GetPane().SetSizer(sizer)
            self.on_pane_changed(None)
        else:
            self.html.set_adr_and_param(mp, None)
            self.on_pane_changed(None)

    def on_pane_changed(self, event):
        if self.html:
            self.GetParent().GetParent().refresh_html()
            self.html.refresh_html()

    def GetBestSize(self):
        ret = wx.CollapsiblePane.GetBestSize(self)
        if self.IsExpanded():
            return (ret[0], ret[1] + self._height)
        elif not self.IsShown():
            return (0, 0)
        else:
            if ret[1] - self._height > 0:
                return (ret[0], ret[1] - self._height)
            else:
                return (ret[0], ret[1])

    def SetValue(self, value):
        if not self.html:
            pass

    def process_refr_data(self, **kwds):
        if "param" in kwds:
            self.param = kwds["param"]
            if "data" in self.param and self.param["data"].strip():
                self.Show(True)
                self.set_html(self.param["data"])


def button_from_parm(parent, param):
    if "children" in param and param["children"]:
        icon = param["children"][0]["attrs"]["class"]
        href = param["attrs"]["href"]
        button = BUTTON(parent, src="fa://" + icon + "?size=0", href=href)
        return button
    else:
        return None


class CompositePanel(wx.Panel, SchBaseCtrl):
    def __init__(self, parent, **kwds):
        SchBaseCtrl.__init__(self, parent, kwds)
        wx.Panel.__init__(self, parent, **kwds)
        self.ctrl = None

    def set_main_ctrl(self, ctrl):
        self.ctrl = ctrl

    def GetValue(self):
        if self.ctrl:
            return str(self.ctrl.GetValue())
        else:
            return None


def SELECT2(parent, **kwds):
    data = kwds["param"]["data"]
    panel = CompositePanel(parent, size=(460, -1))
    ctrl = Select2Base(panel, **kwds)
    panel.set_main_ctrl(ctrl)
    if (
        len(data) > 1
        and "children" in data[1]
        and data[1]["children"]
        and len(data[1]["children"]) > 0
    ):
        button1 = button_from_parm(panel, param=data[1]["children"][0])
    else:
        button1 = None
    if (
        len(data) > 1
        and "children" in data[1]
        and data[1]["children"]
        and len(data[1]["children"]) > 1
    ):
        button2 = button_from_parm(panel, param=data[1]["children"][1])
    else:
        button2 = None

    ctrl.init(button1, button2)

    def ret_ok(id, title):
        ctrl.set_value(id, title)
        wx.CallAfter(ctrl.SetFocus)

    ctrl.ret_ok = ret_ok
    if button1:
        button1.ret_ok = ret_ok
    if button2:
        button2.ret_ok = ret_ok

    box = wx.BoxSizer(wx.HORIZONTAL)
    box.Add(ctrl, 0, wx.EXPAND)
    if button1:
        box.Add(button1)
    if button2:
        box.Add(button2)

    panel.SetSizer(box)
    panel.SetAutoLayout(True)
    panel.Fit()

    return panel


def COMPOSITE(parent, **kwds):
    cls = kwds["param"]["class"].upper()
    if cls in globals():
        return globals()[cls](parent, **kwds)


def COMPONENT(parent, **kwds):
    http = wx.GetApp().get_http(parent)
    response = http.get(
        parent, wx.GetApp().make_href("/schsys/widget_web?browser_type=1")
    )
    buf = response.str()

    colour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE).GetAsString(
        flags=wx.C2S_HTML_SYNTAX
    )
    buf = buf.replace("<component>", kwds["param"]["data"]).replace("[[color]]", colour)

    if HTML2:
        kwds["component"] = True
        obj = HTML2(parent, **kwds)
        obj.load_str(buf, "http://127.0.0.2")
        return obj
    else:
        return None



from autobahn.twisted.websocket import WebSocketClientProtocol, WebSocketClientFactory


class MyClientProtocol(WebSocketClientProtocol):
    def onConnect(self, response):
        print("Server connected: {0}".format(response.peer))

    def onOpen(self):
        print("WebSocket connection open.")

        def hello():
            self.sendMessage("Hello, world!".encode("utf8"))
            self.sendMessage(b"\x00\x01\x03\x04", isBinary=True)
            self.factory.reactor.callLater(1, hello)

        # start sending messages every second ..
        hello()

    def onMessage(self, payload, isBinary):
        if isBinary:
            print("Binary message received: {0} bytes".format(len(payload)))
        else:
            print("Text message received: {0}".format(payload.decode("utf8")))

    def onClose(self, wasClean, code, reason):
        print("WebSocket connection closed: {0}".format(reason))



import wx

from pytigon_gui.guilib.events import *
from pytigon_gui.toolbar.basetoolbar import (
    ToolbarBar,
    ToolbarPage,
    ToolbarPanel,
    ToolbarButton,
)


_ = wx.GetTranslation


class StandardToolbarButton(ToolbarButton):
    def __init__(
        self,
        parent_panel,
        id,
        title,
        bitmap,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        ToolbarButton.__init__(
            self, parent_panel, id, title, bitmap, bitmap_disabled, kind
        )


class StandardToolbarPanel(ToolbarPanel):
    def __init__(self, parent_page, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        ToolbarPanel.__init__(self, parent_page, title, kind)
        if len(parent_page.panels) > 0:
            parent_page.parent_bar.standard_tool_bar.AddSeparator()

    def _append(self, b):
        item = None
        if self.kind in (
            ToolbarPanel.TYPE_PANEL_TOOLBAR,
            ToolbarPanel.TYPE_PANEL_BUTTONBAR,
        ):
            if b.kind == ToolbarButton.TYPE_SIMPLE:
                item = self.parent_page.parent_bar.standard_tool_bar.AddTool(
                    b.id, b.title, b.bitmap, b.title, kind=wx.ITEM_NORMAL
                )
            elif b.kind == ToolbarButton.TYPE_DROPDOWN:
                item = self.parent_page.parent_bar.standard_tool_bar.AddTool(
                    b.id, b.title, b.bitmap, b.title
                )
            elif b.kind == ToolbarButton.TYPE_HYBRID:
                item = self.parent_page.parent_bar.standard_tool_bar.AddTool(
                    b.id, b.title, b.bitmap, b.title
                )
            elif b.kind == ToolbarButton.TYPE_TOOGLE:
                item = self.parent_page.parent_bar.standard_tool_bar.AddToggleTool(
                    b.id, b.title, b.bitmap
                )
            elif b.kind == ToolbarButton.TYPE_PANEL:
                pass
            elif b.kind == ToolbarButton.TYPE_SEPARATOR:
                self.parent_page.parent_bar.standard_tool_bar.AddSeparator()
        else:
            pass

    def create_button(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        b = StandardToolbarButton(self, id, title, bitmap, bitmap_disabled, kind)
        self._append(b)
        return b

    def add_separator(self):
        b = StandardToolbarButton(
            self, 0, "", None, None, kind=ToolbarButton.TYPE_SEPARATOR
        )
        self._append(b)
        return b


class StandardToolbarPage(ToolbarPage):
    def __init__(self, parent_bar, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        ToolbarPage.__init__(self, parent_bar, title, kind)
        if len(parent_bar.pages) > 0:
            parent_bar.standard_tool_bar.AddSeparator()

    def create_panel(self, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        return StandardToolbarPanel(self, title, kind)


class StandardToolbarBar(ToolbarBar):
    def __init__(self, parent, gui_style):
        self.standard_tool_bar = parent.CreateToolBar()
        ToolbarBar.__init__(self, parent, gui_style)

    def create_page(self, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        return StandardToolbarPage(self, title, kind)

    def bind_ui(self, fun, id=wx.ID_ANY):
        self.parent.Bind(wx.EVT_UPDATE_UI, fun, id=id)

    def bind(self, fun, id=wx.ID_ANY, e=None):
        if e:
            self.parent.Bind(e, fun, id=id)
        else:
            self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def bind_dropdown(self, fun, id):
        self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def un_bind(self, id, e=None):
        if e:
            self.Unbind(e, id=id)
        else:
            self.Unbind(wx.EVT_MENU, id=id)



"""Module contains helper classes for button widgets"""

import wx
from wx.lib.agw.ribbon import art
import wx.lib.agw.ribbon as RB
from wx.lib import imageutils


class BitmapTextButton(wx.lib.buttons.GenBitmapButton):
    """Button for toolbars"""

    def __init__(
        self,
        parent,
        id=-1,
        bitmap=wx.NullBitmap,
        label="",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=0,
        validator=wx.DefaultValidator,
        name="bitmaptextbutton",
    ):
        """Constructor"""

        self._art = RB.RibbonMSWArtProvider()
        # self._art = RB.RibbonArtProvider()
        wx.lib.buttons.GenBitmapButton.__init__(
            self, parent, id, bitmap, pos, size, style | wx.BU_EXACTFIT, validator, name
        )
        self.SetLabel(label)
        self.enter_state = False
        self.Bind(wx.EVT_ENTER_WINDOW, self.on_enter)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.on_leave)
        image = self.bmpLabel.ConvertToImage()
        imageutils.grayOut(image)
        self.bmpLabel2 = wx.Bitmap(image)
        self.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE))

    def _GetLabelSize(self):
        s = (self.bmpLabel.GetWidth(), self.bmpLabel.GetHeight())
        ret = self._art.GetButtonBarButtonSize(
            self,
            self,
            art.RIBBON_BUTTON_NORMAL,
            art.RIBBON_BUTTONBAR_BUTTON_LARGE,
            self.GetLabel(),
            s,
            s,
        )
        size_ret = ret[1]
        return (size_ret.GetWidth(), size_ret.GetHeight() - 1, True)

    def DrawLabel(self, dc, width, height, dx=0, dy=0):
        label = self.GetLabel()
        state = art.RIBBON_BUTTONBAR_BUTTON_LARGE
        if self.IsEnabled() and self.GetParent().IsEnabled():
            if self.enter_state:
                state |= art.RIBBON_BUTTONBAR_BUTTON_NORMAL_HOVERED
                if not self.up:
                    state |= art.RIBBON_BUTTONBAR_BUTTON_NORMAL_ACTIVE
            bmp = self.bmpLabel
        else:
            state |= art.RIBBON_BUTTONBAR_BUTTON_DISABLED
            bmp = self.bmpLabel2
        self._art.DrawButtonBarButton(
            dc,
            self,
            wx.Rect(0, 0, width, height),
            art.RIBBON_BUTTON_NORMAL,
            state,
            label,
            bmp,
            bmp,
        )

    def on_enter(self, event):
        if not self.enter_state:
            self.enter_state = True
            wx.CallAfter(self.Refresh)
        event.Skip()

    def on_leave(self, event):
        if self.enter_state:
            self.enter_state = False
            wx.CallAfter(self.Refresh)
        event.Skip()



import wx
from pytigon_gui.guictrl.popup import popuphtml
import string
import datetime
from pdb import set_trace
from pytigon_lib.schtools import schjson

from wx.grid import GridCellEditor


class PopupDataCellControl(popuphtml.DataPopupControl):
    def __init__(self, href, *args, **kwds):
        self.href = href
        self.defaultvalue = None
        popuphtml.DataPopupControl.__init__(self, *args, **kwds)


class PopupDataCellEditor(GridCellEditor):
    def __init__(self):
        GridCellEditor.__init__(self)
        self.address = None
        self.param = None

    def get_address(self):
        return self._address

    def set_address(self, address):
        if address != None and address[0] == "!":
            set_trace()
        self._address = address

    address = property(get_address, set_address)

    def Create(self, parent, id, evt_handler):
        self.Parent = parent
        self._tc = PopupDataCellControl(self.address, parent, id)
        self._tc.DismissObject = self
        self.SetControl(self._tc)
        self.evtHandler = evt_handler

        if evt_handler:
            self._tc.PushEventHandler(self.evtHandler)
            evt_handler.Bind(wx.EVT_KILL_FOCUS, self.on_kill_focus)

    def on_kill_focus(self, event):
        pass

    def SetSize(self, rect):
        self._tc.SetSize(
            rect.x, rect.y, rect.width + 2, rect.height + 2, wx.SIZE_ALLOW_MINUS_ONE
        )

    def Show(self, show, attr):
        GridCellEditor.Show(self, show, attr)

    def PaintBackground(self, rect, attr):
        pass

    def BeginEdit(self, row, col, grid):
        self.grid = grid
        self.start_value = None
        value = grid.GetTable().GetValue(row, col)

        if value.__class__ == datetime.date:
            self.start_value = value.isoformat()

        if type(value) == str:
            self.start_value = value
        else:
            self.start_value = str(value)

        self._tc.focus_in(self.start_value)
        if self.start_value:
            self._tc.set_rec(self.start_value, (self.start_value,))
        self._tc.SetInsertionPointEnd()
        self._tc.GetTextCtrl().SetSelection(0, -1)
        self._tc.GetTextCtrl().SetFocus()

    def EndEdit(self, row, col, grid, old_value):
        self._tc.focus_out(self._tc.GetValue())
        changed = False
        value = self._tc.GetValue()
        if str(value) != str(self.start_value):
            changed = True
            grid.GetTable().SetValue(row, col, value)  # update the table
        grid.SetFocus()
        return changed

    def ApplyEdit(self, row, col, grid):
        val = self._tc.GetValue()
        grid.GetTable().SetValue(row, col, val)
        self.start_value = ""
        self._tc.SetValue("")

    def Reset(self):
        self._tc.SetValue(self.start_value)

    def IsAcceptedKey(self, evt):
        return (
            not (evt.ControlDown() or evt.AltDown())
            and evt.GetKeyCode() != wx.WXK_SHIFT
        )

    def StartingKey(self, evt):
        key = evt.GetKeyCode()
        ch = None
        if key < 256 and key >= 0 and chr(key) in string.printable:
            ch = chr(key)
        if ch is not None:
            self._tc.start_value = ""
            self._tc.SetValue(ch)
            if self._tc.popup:
                self._tc.popup.set_string_value(ch)
            self._tc.rec_value = (ch,)
            if self._tc.popup:
                self._tc.popup.Dismiss()
            self._tc.SetInsertionPointEnd()
        else:
            evt.Skip()

    def StartingClick(self):
        print("START CLICK")
        pass

    def Clone(self):
        ret = PopupDataCellEditor()
        ret.SetParameters(self.param)
        return ret

    def set_parameters(self, params):
        self.param = params


class DatePopupDataCellEditor(PopupDataCellEditor):
    def __init__(self):
        PopupDataCellEditor.__init__(self)
        self.address = wx.GetApp().make_href("/schsys/datedialog/")

    def Create(
        self,
        parent,
        id,
        evt_handler,
    ):
        PopupDataCellEditor.Create(self, parent, id, evt_handler)
        self._tc.to_masked(autoformat="EUDATEYYYYMMDD.")

    def Clone(self):
        ret = DatePopupDataCellEditor()
        ret.address = self.address
        ret.SetParameters(self.param)
        return ret


class ListPopupCellEditor(PopupDataCellEditor):
    def __init__(self):
        PopupDataCellEditor.__init__(self)
        self.address = wx.GetApp().make_href("/schsys/listdialog/")

    def Create(self, parent, id, evt_handler):
        PopupDataCellEditor.Create(self, parent, id, evt_handler)
        self._tc.SetEventObject(self)

    def Clone(self):
        ret = ListPopupCellEditor()
        ret.address = self.address
        ret.SetParameters(self.param)
        return ret

    def BeginEdit(self, row, col, grid):
        typ = grid.GetTable().GetTypeName(row, col)
        id = typ.find(":")
        self.choices = schjson.loads(typ[id + 1 :])
        PopupDataCellEditor.BeginEdit(self, row, col, grid)

    def OnButtonClick(self):
        if self._tc.simpleDialog:
            self._tc.popup.html.choices = self.choices
            self._tc.popup.html.refr()
        else:
            self._tc.page.body.choices = self.choices
            self._tc.page.body.refr()

    def set_rec(self, value, value_rec, dismiss=True):
        ret = ""
        if len(value) > 2 and value[1] == ":":
            return value
        else:
            for choice in self.choices:
                if choice[1].lower().startswith(value.lower()):
                    ret = choice[0] + ":" + choice[1]
            return ret


class GenericPopupCellEditor(PopupDataCellEditor):
    def __init__(self):
        PopupDataCellEditor.__init__(self)
        self.address = wx.GetApp().make_href("/schsys/datedialog/")

    def Clone(self):
        ret = GenericPopupCellEditor()
        ret.SetParameters(self.address)
        return ret

    def set_parameters(self, params):
        self.address = params



"""
Module contains base classes for toolbars and menu:
Class hierary:
ToolbarBar
    ToolbarPage
        ToolbarPanel
            ToolbarButton
            ToolbarButton
            ...
        ToolbarPanel
        ....
        BaseHtmlPanel
    ....

One ToolbarBar can contains many ToolbarPage objects, one ToolbarPage can contains many ToolbarPanel objects etc.

There are real toolbars and menus, which are base on those abstract classes:
- MenuToolbarBar
- GenericToolbarBar
- StandardToolbarBar
- ModernToolbarBar
- TreeToolbarBar
"""

import wx

from pytigon_gui.guilib.events import *
from pytigon_gui.toolbar.standardtoolbarbuttons import StandardButtons
from pytigon_gui.guiframe.page import SchPage

_ = wx.GetTranslation


class BaseHtmlPanel:
    def __init__(self, page, real_panel):
        self.page = page
        self.real_panel = real_panel
        self.html_page = None
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.real_panel.SetSizer(self.sizer)

    def get_width(self):
        return 0

    def get_height(self):
        return 0

    def get_window(self):
        return self.real_panel

    def set_page(self, html_page):
        if self.html_page:
            self.sizer.Replace(self.html_page, html_page)
            self.html_page.Destroy()
        else:
            self.sizer.Add(html_page, 0, wx.LEFT | wx.TOP | wx.EXPAND | wx.RIGHT, 2)

        self.html_page = html_page

        self.sizer.Fit(self.real_panel)


class ToolbarButton:
    """Toolbar button"""

    TYPE_SIMPLE = 0
    TYPE_DROPDOWN = 1
    TYPE_HYBRID = 2
    TYPE_TOOGLE = 3
    TYPE_PANEL = 4
    TYPE_SEPARATOR = 5

    def __init__(
        self,
        parent_panel,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=TYPE_SIMPLE,
    ):
        """Constructor, creating a new toolbar button

        Args:
            parent_panel - parent ToolbarPanel object
            id - An integer by which the tool may be identified
            title - button title
            bitmap - button bitmap
            bitmap_disabled - button bitmap used when button is disabled
            kind - type of button, may be:
                TYPE_SIMPLE = 0
                TYPE_DROPDOWN = 1
                TYPE_HYBRID = 2
                TYPE_TOOGLE = 3
                TYPE_PANEL = 4
                TYPE_SEPARATOR = 5
        """
        self.parent_panel = parent_panel
        self.id = id
        self.title = title
        self.bitmap = bitmap
        if bitmap_disabled is None and bitmap is not None:
            self.bitmap_disabled = wx.NullBitmap
        else:
            self.bitmap_disabled = bitmap_disabled
        self.kind = kind


class ToolbarPanel(object):
    """Toolbar panel"""

    TYPE_PANEL_BUTTONBAR = 0
    TYPE_PANEL_TOOLBAR = 1
    TYPE_PANEL_PANELBAR = 2

    def __init__(self, parent_page, title, kind=TYPE_PANEL_BUTTONBAR):
        """Constructor, creating a new toolbar panel

        Args:
            parent_page - parent ToolbarPage object
            title - panel title
            kind - type of panel, may be:
                TYPE_PANEL_BUTTONBAR = 0
                TYPE_PANEL_TOOLBAR = 1
                TYPE_PANEL_PANELBAR = 2
        """
        self.parent_page = parent_page
        self.title = title
        self.kind = kind
        self.buttons = []

    def _transform_bitmaps_parm(self, bitmaps):
        b = [None, None]
        if len(bitmaps) > 0:
            b[0] = bitmaps[0]
        if len(bitmaps) > 1:
            b[1] = bitmaps[1]
        return b

    def create_button(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        """Function return new ToolbarButton, shoud be derived. In derived class shoud return ToolbarButton
        derived object.

        Args: see ToolbarButton contructor
        """
        return ToolbarButton(self, id, title, bitmap, bitmap_disabled, kind)

    def append(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        """Append button to panel.

        Args: see ToolbarButton contructor
        """
        button = self.create_button(id, title, bitmap, bitmap_disabled, kind)
        self.buttons.append(button)
        return button

    def add_simple_tool(self, id, title, bitmaps):
        """Append simple tool to toolbar.

        Args:
            id - an integer by which the tool may be identified
            title - button title
            bitmaps - list of bitmaps, can contains 0, 1 or 2 bitmaps
                first bitmap is used for tool in normal state, second for disabled tool.
        """
        b = self._transform_bitmaps_parm(bitmaps)
        return self.append(id, title, b[0], b[1], kind=ToolbarButton.TYPE_SIMPLE)

    def add_dropdown_tool(self, id, title, bitmaps):
        """Append simple tool to toolbar.

        Args:
            id - an integer by which the tool may be identified
            title - button title
            bitmaps - list of bitmaps, can contains 0, 1 or 2 bitmaps
                first bitmap is used for tool in normal state, second for disabled tool.
        """
        b = self._transform_bitmaps_parm(bitmaps)
        return self.append(id, title, b[0], b[1], kind=ToolbarButton.TYPE_DROPDOWN)

    def add_hybrid_tool(self, id, title, bitmaps):
        """Append hybrid tool to toolbar.

        Args:
            id - an integer by which the tool may be identified
            title - button title
            bitmaps - list of bitmaps, can contains 0, 1 or 2 bitmaps
                first bitmap is used for tool in normal state, second for disabled tool.
        """
        b = self._transform_bitmaps_parm(bitmaps)
        return self.append(id, title, b[0], b[1], kind=ToolbarButton.TYPE_HYBRID)

    def add_toogle_tool(self, id, title, bitmaps):
        """Append toogle tool to toolbar.

        Args:
            id - an integer by which the tool may be identified
            title - button title
            bitmaps - list of bitmaps, can contains 0, 1 or 2 bitmaps
                first bitmap is used for tool in normal state, second for disabled tool.
        """
        b = self._transform_bitmaps_parm(bitmaps)
        return self.append(id, title, b[0], b[1], kind=ToolbarButton.TYPE_TOOGLE)

    def add_separator(self):
        """Append separator to toolbar."""
        pass


class ToolbarPage(object):
    """Toolbar page"""

    TYPE_PAGE_NORMAL = 0

    def __init__(self, parent_bar, title, kind=TYPE_PAGE_NORMAL):
        """Constructor, creating a new ToolbarPage page.

        Args:
            parent_bar - parent ToolbarBar object
            title - page title
            kind - type of page, may be:
                TYPE_PAGE_NORMAL = 0
        """
        self.parent_bar = parent_bar
        self.name = title
        self.title = _(title)
        self.kind = kind
        self.panels = []

    def create_panel(self, title, kind):
        """Function return new ToolbarPanel, shoud be derived. In derived class shoud return ToolbarPanel
        derived object.

        Args: see ToolbarPanel contructor
        """
        return self.ToolbarPanel(self, title, kind=ToolbarPanel.TYPE_PANEL_BUTTONBAR)

    def create_html_panel(self, title):
        """Function shoud be overwriten in derived class.
        In derived class shoud return BaseHtmlPanel derived object.

        Args: see BaseHtmlPanel contructor
        """
        return None

    def append(self, title, kind=ToolbarPanel.TYPE_PANEL_BUTTONBAR):
        """Append new ToolbarPanel to page.

        Args:
            title - panel title
            kind - see ToolbarPanel constructor

        """
        p = self.create_panel(title, kind)
        self.panels.append(p)
        return p


class ToolbarBar(object):
    """Toolbar bar"""

    def __init__(self, parent, gui_style):
        """Constructor, ToolbarBar is base class for all menus and toolbars connected to top frame window.

        Args:
            parent - parent window - top wx.Frame derived object
            gui_style - string description of gui interface. Base on this string standard toolbar elements are created.
        """
        self.parent = parent
        self.main_page = None
        self.gui_style = gui_style

        self.toolbars = {}
        self.pages = []
        self.user_panels = {}

        self.standard_buttons = StandardButtons(self, gui_style)
        self.standard_buttons.create_file_panel(self.main_page)
        self.standard_buttons.create_edit_panel(self.main_page)
        self.standard_buttons.create_operations_panel(self.main_page)
        self.standard_buttons.create_browse_panel(self.main_page)
        self.standard_buttons.create_address_panel(self.main_page)

    def create_page(self, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        """Function return new ToolbarPage, shoud be derived. In derived class shoud return ToolbarPage
        derived object.

        Args: see ToolbarPage contructor
        """
        return self.ToolbarPage(self, title, kind)

    def remove_page(self, title):
        """Remove page witch specified title from toolbar

        Args:
            title: title of removed page
        """
        for page in self.pages:
            if page.title == title:
                self.pages.remove(page)
                break
        for key, panel in self.user_panels.items():
            if panel.page.title == title:
                del self.user_panels[key]
                return

    def append(self, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        """Append page to toolbar.

        Args: see ToolbarPage constructor
        """
        for page in self.pages:
            if page.name == title and page.kind == kind:
                return page
        p = self.create_page(title, kind)
        if not self.main_page:
            self.main_page = p
        self.pages.append(p)
        return p

    def create(self):
        """Create toolbar"""
        pass

    def close(self):
        """Close toolbar"""
        pass

    def create_panel_in_main_page(self, title, kind):
        """Create panel in a main page

        Args:
            title - new panels title
            kind - type of new panel - see ToolbarPanel constructor.
        """
        if not self.main_page:
            self.append("main tools")
        return self.main_page.create_panel(title, kind)

    def bind(self, fun, id=wx.ID_ANY, e=None):
        """Bind event handler to toolbar

        Args:
            fun - handler function
            id - identifier of event
            e - type of event
        """
        if e:
            self.parent.Bind(e, fun, id=id)
        else:
            self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def un_bind(self, id=wx.ID_ANY, e=None):
        """Unbind event handler from toolbar."""
        if e:
            self.parent.Unbind(e, id=id)
        else:
            self.parent.Unbind(wx.EVT_MENU, id=id)

    def create_html_win(
        self, toolbar_page, address_or_parser, parameters, callback=None
    ):
        """Create html page in toolbar. Not all toolbars may be used

        Args:
            toolbar_page - name of toolbar page
            address_or_parser - addres of http page
            parameters - parameters for http request
        """
        page = None
        if not toolbar_page:
            u_name = "main tools"
            page_name = _("main tools")
            panel_name = "main tools"
        else:
            u_name = toolbar_page
            names = toolbar_page.split("__")
            if len(names) > 1:
                page_name = names[0].replace("_", " ")
                panel_name = names[1].replace("_", " ")
            else:
                page_name = toolbar_page.replace("_", " ")
                panel_name = page_name

        if u_name in self.user_panels:
            panel = self.user_panels[u_name]
        else:
            if page_name in self.pages:
                page = self.pages[page]
            else:
                page = self.append(page_name)
            panel = page.create_html_panel(panel_name)

        if panel:
            dx = panel.get_width() + 3
            dy = panel.get_height() + 5

            page2 = SchPage(
                panel.get_window(),
                address_or_parser,
                parameters,
                size=wx.Size(dx, dy),
                pos=wx.Point(2, 2),
            )
            best = page2.body.calculate_best_size()
            page2.SetSize(wx.Size(int(best[0]), int(best[1])))
            panel.set_page(page2)
            self.user_panels[u_name] = panel
            page2.body.toolbar_interface = self
            page2.body.toolbar_interface_page = page

            def init_page(callback):
                page2.init_frame()
                page2.activate_page()
                # wx.GetApp().GetTopWindow()._mgr.GetPane("desktop").Show()
                page2.Update()

            wx.CallAfter(init_page, callback)

            return page2

        return None

    def new_child_page(self, address_or_parser, title="", param=None, callback=None):
        """Crate new chid page, for toolbar child page is transfered to desktop window"""
        return wx.GetApp().GetTopWindow().new_main_page(address_or_parser, title, param)



"""
Hierarchia okien aplikacji:
  appframe.SChAppFrame
    self.Desktop = appframe.SchNotebook
      appframe.SchNotebook
        appnotebookpage.SchNotebookPage
            SChSashWindow (spełnia funkcję ParmObj)
              self.header = SChFrame
              self.body = SChFrame
              self.footer = SChFrame
              self.panel = SChFrame
"""
